import logging
import os
import shlex

import borgmatic.borg.pattern
import borgmatic.config.paths
import borgmatic.hooks.credential.parse
import borgmatic.hooks.data_source.config
from borgmatic.execute import execute_command, execute_command_with_processes
from borgmatic.hooks.data_source import config as database_config
from borgmatic.hooks.data_source import dump

logger = logging.getLogger(__name__)


def make_dump_path(base_directory):  # pragma: no cover
    '''
    Given a base directory, make the corresponding dump path.
    '''
    return dump.make_data_source_dump_path(base_directory, 'mongodb_databases')


def get_default_port(databases, config):  # pragma: no cover
    return 27017


def use_streaming(databases, config):
    '''
    Given a sequence of MongoDB database configuration dicts, a configuration dict (ignored), return
    whether streaming will be using during dumps.
    '''
    return any(database.get('format') != 'directory' for database in databases)


def dump_data_sources(
    databases,
    config,
    config_paths,
    borgmatic_runtime_directory,
    patterns,
    dry_run,
):
    '''
    Dump the given MongoDB databases to a named pipe. The databases are supplied as a sequence of
    dicts, one dict describing each database as per the configuration schema. Use the borgmatic
    runtime directory to construct the destination path (used for the directory format.

    Return a sequence of subprocess.Popen instances for the dump processes ready to spew to a named
    pipe. But if this is a dry run, then don't actually dump anything and return an empty sequence.
    Also append the the parent directory of the database dumps to the given patterns list, so the
    dumps actually get backed up.
    '''
    dry_run_label = ' (dry run; not actually dumping anything)' if dry_run else ''

    logger.info(f'Dumping MongoDB databases{dry_run_label}')

    processes = []
    dumps_metadata = []

    for database in databases:
        name = database['name']
        dumps_metadata.append(
            borgmatic.actions.restore.Dump(
                'mongodb_databases',
                name,
                database.get('hostname'),
                database.get('port'),
                database.get('label'),
                database.get('container'),
            )
        )

        dump_filename = dump.make_data_source_dump_filename(
            make_dump_path(borgmatic_runtime_directory),
            name,
            hostname=database.get('hostname'),
            port=database.get('port'),
            container=database.get('container'),
            label=database.get('label'),
        )
        dump_format = database.get('format', 'archive')

        logger.debug(
            f'Dumping MongoDB database {name} to {dump_filename}{dry_run_label}',
        )
        if dry_run:
            continue

        command = build_dump_command(database, config, dump_filename, dump_format)

        if dump_format == 'directory':
            dump.create_parent_directory_for_dump(dump_filename)
            execute_command(  # noqa: S604
                command,
                shell=True,
                working_directory=borgmatic.config.paths.get_working_directory(config),
            )
        else:
            dump.create_named_pipe_for_dump(dump_filename)
            processes.append(
                execute_command(  # noqa: S604
                    command,
                    shell=True,
                    run_to_completion=False,
                    working_directory=borgmatic.config.paths.get_working_directory(config),
                ),
            )

    if not dry_run:
        dump.write_data_source_dumps_metadata(
            borgmatic_runtime_directory, 'mongodb_databases', dumps_metadata
        )
        borgmatic.hooks.data_source.config.inject_pattern(
            patterns,
            borgmatic.borg.pattern.Pattern(
                os.path.join(borgmatic_runtime_directory, 'mongodb_databases'),
                source=borgmatic.borg.pattern.Pattern_source.HOOK,
            ),
        )

    return processes


def make_password_config_file(password):
    '''
    Given a database password, write it as a MongoDB configuration file to an anonymous pipe and
    return its filename. The idea is that this is a more secure way to transmit a password to
    MongoDB than providing it directly on the command-line.

    Do not use the returned value for multiple different command invocations. That will not work
    because each pipe is "used up" once read.
    '''
    logger.debug('Writing MongoDB password to configuration file pipe')

    read_file_descriptor, write_file_descriptor = os.pipe()
    os.write(write_file_descriptor, f'password: {password}'.encode())
    os.close(write_file_descriptor)

    # This plus subprocess.Popen(..., close_fds=False) in execute.py is necessary for the database
    # client child process to inherit the file descriptor.
    os.set_inheritable(read_file_descriptor, True)

    return f'/dev/fd/{read_file_descriptor}'


def build_dump_command(database, config, dump_filename, dump_format):
    '''
    Return the custom mongodump_command from a single database configuration.
    '''
    all_databases = database['name'] == 'all'

    password = borgmatic.hooks.credential.parse.resolve_credential(database.get('password'), config)

    dump_command = tuple(
        shlex.quote(part) for part in shlex.split(database.get('mongodump_command') or 'mongodump')
    )
    hostname = database_config.resolve_database_option('hostname', database)
    return (
        dump_command
        + (('--out', shlex.quote(dump_filename)) if dump_format == 'directory' else ())
        + (('--host', shlex.quote(hostname)) if hostname else ())
        + (('--port', shlex.quote(str(database['port']))) if 'port' in database else ())
        + (
            (
                '--username',
                shlex.quote(
                    borgmatic.hooks.credential.parse.resolve_credential(
                        database['username'],
                        config,
                    ),
                ),
            )
            if 'username' in database
            else ()
        )
        + (('--config', make_password_config_file(password)) if password else ())
        + (
            (
                '--authenticationDatabase',
                shlex.quote(database['authentication_database']),
            )
            if 'authentication_database' in database
            else ()
        )
        + (('--db', shlex.quote(database['name'])) if not all_databases else ())
        + (
            tuple(shlex.quote(option) for option in database['options'].split(' '))
            if 'options' in database
            else ()
        )
        + (('--archive', '>', shlex.quote(dump_filename)) if dump_format != 'directory' else ())
    )


def remove_data_source_dumps(
    databases,
    config,
    borgmatic_runtime_directory,
    patterns,
    dry_run,
):  # pragma: no cover
    '''
    Remove all database dump files for this hook regardless of the given databases. Use the
    borgmatic_runtime_directory to construct the destination path. If this is a dry run, then don't
    actually remove anything.
    '''
    dump.remove_data_source_dumps(make_dump_path(borgmatic_runtime_directory), 'MongoDB', dry_run)


def make_data_source_dump_patterns(
    databases,
    config,
    borgmatic_runtime_directory,
    name=None,
    hostname=None,
    port=None,
    container=None,
    label=None,
):  # pragma: no cover
    '''
    Given a sequence of configurations dicts, a configuration dict, the borgmatic runtime directory,
    and a database name to match, return the corresponding glob patterns to match the database dump
    in an archive.
    '''
    borgmatic_source_directory = borgmatic.config.paths.get_borgmatic_source_directory(config)

    return (
        dump.make_data_source_dump_filename(
            make_dump_path('borgmatic'), name, hostname, port, container, label
        ),
        dump.make_data_source_dump_filename(
            make_dump_path(borgmatic_runtime_directory),
            name,
            hostname,
            port,
            container,
            label,
        ),
        dump.make_data_source_dump_filename(
            make_dump_path(borgmatic_source_directory),
            name,
            hostname,
            port,
            container,
            label,
        ),
    )


def restore_data_source_dump(
    hook_config,
    config,
    data_source,
    dry_run,
    extract_process,
    connection_params,
    borgmatic_runtime_directory,
):
    '''
    Restore a database from the given extract stream. The database is supplied as a data source
    configuration dict, but the given hook configuration is ignored. The given configuration dict is
    used to construct the destination path. If this is a dry run, then don't actually restore
    anything. Trigger the given active extract process (an instance of subprocess.Popen) to produce
    output to consume.

    If the extract process is None, then restore the dump from the filesystem rather than from an
    extract stream.
    '''
    dry_run_label = ' (dry run; not actually restoring anything)' if dry_run else ''
    dump_filename = dump.make_data_source_dump_filename(
        make_dump_path(borgmatic_runtime_directory),
        data_source['name'],
        hostname=data_source.get('hostname'),
        port=data_source.get('port'),
        container=data_source.get('container'),
        label=data_source.get('label'),
    )
    restore_command = build_restore_command(
        extract_process,
        data_source,
        config,
        dump_filename,
        connection_params,
    )

    logger.debug(f'Restoring MongoDB database {data_source["name"]}{dry_run_label}')
    if dry_run:
        return

    # Don't give Borg local path so as to error on warnings, as "borg extract" only gives a warning
    # if the restore paths don't exist in the archive.
    tuple(
        execute_command_with_processes(
            restore_command,
            [extract_process] if extract_process else [],
            output_log_level=logging.DEBUG,
            input_file=extract_process.stdout if extract_process else None,
            working_directory=borgmatic.config.paths.get_working_directory(config),
            borg_local_path=config.get('local_path', 'borg'),
        )
    )


def build_restore_command(extract_process, database, config, dump_filename, connection_params):
    '''
    Return the custom mongorestore_command from a single database configuration.
    '''
    hostname = database_config.resolve_database_option(
        'hostname', database, connection_params, restore=True
    )
    port = database_config.resolve_database_option(
        'port', database, connection_params, restore=True
    )
    username = borgmatic.hooks.credential.parse.resolve_credential(
        database_config.resolve_database_option(
            'username', database, connection_params, restore=True
        ),
        config,
    )
    password = borgmatic.hooks.credential.parse.resolve_credential(
        database_config.resolve_database_option(
            'password', database, connection_params, restore=True
        ),
        config,
    )

    command = [
        shlex.quote(part)
        for part in shlex.split(database.get('mongorestore_command') or 'mongorestore')
    ]

    if extract_process:
        command.append('--archive')
    else:
        command.extend(('--dir', dump_filename))

    if database['name'] != 'all':
        command.extend(('--drop',))

    if hostname:
        command.extend(('--host', hostname))

    if port:
        command.extend(('--port', str(port)))

    if username:
        command.extend(('--username', username))

    if password:
        command.extend(('--config', make_password_config_file(password)))

    if 'authentication_database' in database:
        command.extend(('--authenticationDatabase', database['authentication_database']))

    if 'restore_options' in database:
        command.extend(database['restore_options'].split(' '))

    if database.get('schemas'):
        for schema in database['schemas']:
            command.extend(('--nsInclude', schema))

    return command



import logging

import borgmatic.borg.export_key

logger = logging.getLogger(__name__)


def run_export_key(
    repository,
    config,
    local_borg_version,
    export_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "key export" action for the given repository.
    '''
    logger.info('Exporting repository key')
    borgmatic.borg.export_key.export_key(
        repository['path'],
        config,
        local_borg_version,
        export_arguments,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )



import logging
import shlex

import borgmatic.borg.environment
import borgmatic.borg.flags

logger = logging.getLogger(__name__)


def make_rename_command(
    dry_run,
    repository_name,
    old_archive_name,
    new_archive_name,
    config,
    local_borg_version,
    local_path,
    remote_path,
):
    extra_borg_options = config.get('extra_borg_options', {}).get('rename', '')

    return (
        (local_path, 'rename')
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + borgmatic.borg.flags.make_flags('dry-run', dry_run)
        + borgmatic.borg.flags.make_flags('remote-path', remote_path)
        + borgmatic.borg.flags.make_flags('umask', config.get('umask'))
        + ('--log-json',)
        + borgmatic.borg.flags.make_flags('lock-wait', config.get('lock_wait'))
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + borgmatic.borg.flags.make_repository_archive_flags(
            repository_name,
            old_archive_name,
            local_borg_version,
        )
        + (new_archive_name,)
    )


def rename_archive(
    repository_name,
    old_archive_name,
    new_archive_name,
    dry_run,
    config,
    local_borg_version,
    local_path,
    remote_path,
):
    command = make_rename_command(
        dry_run,
        repository_name,
        old_archive_name,
        new_archive_name,
        config,
        local_borg_version,
        local_path,
        remote_path,
    )

    borgmatic.execute.execute_command(
        command,
        output_log_level=logging.INFO,
        environment=borgmatic.borg.environment.make_environment(config),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import collections
import enum
import logging
import os
import tempfile

logger = logging.getLogger(__name__)


# See https://borgbackup.readthedocs.io/en/stable/usage/help.html#borg-help-patterns
class Pattern_type(enum.Enum):
    ROOT = 'R'  # A ROOT pattern always has a NONE pattern style.
    PATTERN_STYLE = 'P'
    EXCLUDE = '-'
    NO_RECURSE = '!'
    INCLUDE = '+'


class Pattern_style(enum.Enum):
    NONE = ''
    FNMATCH = 'fm'
    SHELL = 'sh'
    REGULAR_EXPRESSION = 're'
    PATH_PREFIX = 'pp'
    PATH_FULL_MATCH = 'pf'


class Pattern_source(enum.Enum):
    '''
    Where the pattern came from within borgmatic. This is important because certain use cases (like
    filesystem snapshotting) only want to consider patterns that the user actually put in a
    configuration file and not patterns from other sources.
    '''

    # The pattern is from a borgmatic configuration option, e.g. listed in "source_directories".
    CONFIG = 'config'

    # The pattern is generated internally within borgmatic, e.g. for special file excludes.
    INTERNAL = 'internal'

    # The pattern originates from within a borgmatic hook, e.g. a database hook that adds its dump
    # directory.
    HOOK = 'hook'


Pattern = collections.namedtuple(
    'Pattern',
    ('path', 'type', 'style', 'device', 'source'),
    defaults=(
        Pattern_type.ROOT,
        Pattern_style.NONE,
        None,
        Pattern_source.HOOK,
    ),
)


def write_patterns_file(patterns, borgmatic_runtime_directory, patterns_file=None):
    '''
    Given a sequence of patterns as Pattern instances, write them to a named temporary file in the
    given borgmatic runtime directory and return the file object so it can continue to exist on disk
    as long as the caller needs it.

    If an optional open pattern file is given, append to it instead of making a new temporary file.
    Return None if no patterns are provided.
    '''
    if not patterns:
        return None

    if patterns_file is None:
        patterns_file = tempfile.NamedTemporaryFile(
            'w', dir=borgmatic_runtime_directory, encoding='utf-8'
        )
        operation_name = 'Writing'
    else:
        patterns_file.write('\n')
        operation_name = 'Appending'

    patterns_output = '\n'.join(
        f'{pattern.type.value} {pattern.style.value}{":" if pattern.style.value else ""}{pattern.path}'
        for pattern in patterns
    )
    logger.debug(f'{operation_name} patterns to {patterns_file.name}:\n{patterns_output}')

    patterns_file.write(patterns_output)
    patterns_file.flush()

    return patterns_file


def check_all_root_patterns_exist(patterns):
    '''
    Given a sequence of Pattern instances, check that all root pattern paths exist. If any don't,
    raise an exception.
    '''
    missing_paths = [
        pattern.path
        for pattern in patterns
        if pattern.type == Pattern_type.ROOT
        if not os.path.exists(pattern.path)
    ]

    if missing_paths:
        raise ValueError(
            f"Source directories or root pattern paths do not exist: {', '.join(missing_paths)}",
        )



import datetime
import json
import logging
import platform

import requests

import borgmatic.hooks.credential.parse
import borgmatic.hooks.monitoring.logs
from borgmatic.hooks.monitoring import monitor

logger = logging.getLogger(__name__)

EVENTS_API_URL = 'https://events.pagerduty.com/v2/enqueue'
DEFAULT_LOGS_PAYLOAD_LIMIT_BYTES = 10000
HANDLER_IDENTIFIER = 'pagerduty'
TIMEOUT_SECONDS = 10


def initialize_monitor(hook_config, config, config_filename, monitoring_log_level, dry_run):
    '''
    Add a handler to the root logger that stores in memory the most recent logs emitted. That way,
    we can send them all to PagerDuty upon a failure state. But skip this if the "send_logs" option
    is false.
    '''
    if hook_config.get('send_logs', False) is False:
        return

    ping_body_limit = max(
        DEFAULT_LOGS_PAYLOAD_LIMIT_BYTES
        - len(borgmatic.hooks.monitoring.logs.PAYLOAD_TRUNCATION_INDICATOR),
        0,
    )

    borgmatic.hooks.monitoring.logs.add_handler(
        borgmatic.hooks.monitoring.logs.Forgetful_buffering_handler(
            HANDLER_IDENTIFIER,
            ping_body_limit,
            monitoring_log_level,
        ),
    )


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    If this is an error state, create a PagerDuty event with the configured integration key. Use
    the given configuration filename in any log entries. If this is a dry run, then don't actually
    create an event.
    '''
    if state != monitor.State.FAIL:
        logger.debug(
            f'Ignoring unsupported monitoring state {state.name.lower()} in PagerDuty hook',
        )
        return

    dry_run_label = ' (dry run; not actually sending)' if dry_run else ''
    logger.info(f'Sending failure event to PagerDuty {dry_run_label}')

    try:
        integration_key = borgmatic.hooks.credential.parse.resolve_credential(
            hook_config.get('integration_key'),
            config,
        )
    except ValueError as error:
        logger.warning(f'PagerDuty credential error: {error}')
        return

    logs_payload = borgmatic.hooks.monitoring.logs.format_buffered_logs_for_payload(
        HANDLER_IDENTIFIER,
    )

    hostname = platform.node()
    local_timestamp = datetime.datetime.now(datetime.timezone.utc).astimezone().isoformat()
    payload = json.dumps(
        {
            'routing_key': integration_key,
            'event_action': 'trigger',
            'payload': {
                'summary': f'backup failed on {hostname}',
                'severity': 'error',
                'source': hostname,
                'timestamp': local_timestamp,
                'component': 'borgmatic',
                'group': 'backups',
                'class': 'backup failure',
                'custom_details': {
                    'hostname': hostname,
                    'configuration filename': config_filename,
                    'server time': local_timestamp,
                    'logs': logs_payload,
                },
            },
        },
    )

    if dry_run:
        return

    logging.getLogger('urllib3').setLevel(logging.ERROR)
    try:
        response = requests.post(
            EVENTS_API_URL,
            data=payload.encode('utf-8'),
            timeout=TIMEOUT_SECONDS,
            headers={'User-Agent': 'borgmatic'},
        )
        if not response.ok:
            response.raise_for_status()
    except requests.exceptions.RequestException as error:
        logger.warning(f'PagerDuty error: {error}')


def destroy_monitor(ping_url_or_uuid, config, monitoring_log_level, dry_run):  # pragma: no cover
    '''
    Remove the monitor handler that was added to the root logger. This prevents the handler from
    getting reused by other instances of this monitor.
    '''
    borgmatic.hooks.monitoring.logs.remove_handler(HANDLER_IDENTIFIER)



import enum
import json
import logging
import logging.handlers
import os
import socket
import sys


def to_bool(arg):
    '''
    Return a boolean value based on `arg`.
    '''
    if arg is None or isinstance(arg, bool):
        return arg

    if isinstance(arg, str):
        arg = arg.lower()

    return arg in {'yes', 'on', '1', 'true', 1}


def interactive_console():
    '''
    Return whether the current console is "interactive". Meaning: Capable of
    user input and not just something like a cron job.
    '''
    return sys.stderr.isatty() and os.environ.get('TERM') != 'dumb'


def should_do_markup(configs, json_enabled):
    '''
    Given a dict of configuration filename to corresponding parsed configuration (which already have
    any command-line overrides applied) and whether json is enabled, determine if we should enable
    color marking up.
    '''
    if json_enabled:
        return False

    if any(config.get('color', True) is False for config in configs.values()):
        return False

    if os.environ.get('NO_COLOR', None):
        return False

    py_colors = os.environ.get('PY_COLORS', None)

    if py_colors is not None:
        return to_bool(py_colors)

    return interactive_console()


class Multi_stream_handler(logging.Handler):
    '''
    A logging handler that dispatches each log record to one of multiple stream handlers depending
    on the record's log level.
    '''

    def __init__(self, log_level_to_stream_handler):
        super().__init__()
        self.log_level_to_handler = log_level_to_stream_handler
        self.handlers = set(self.log_level_to_handler.values())

    def flush(self):  # pragma: no cover
        super().flush()

        for handler in self.handlers:
            handler.flush()

    def emit(self, record):
        '''
        Dispatch the log record to the appropriate stream handler for the record's log level.
        '''
        self.log_level_to_handler[record.levelno].emit(record)

    def setFormatter(self, formatter):  # pragma: no cover  # noqa: N802
        super().setFormatter(formatter)

        for handler in self.handlers:
            handler.setFormatter(formatter)

    def setLevel(self, level):  # pragma: no cover  # noqa: N802
        super().setLevel(level)

        for handler in self.handlers:
            handler.setLevel(level)


DEFAULT_JOURNALD_PRIORITY = 6


class JournaldHandler(logging.Handler):
    def __init__(self, journald_socket_path):
        super().__init__()

        add_custom_log_levels()

        self.journald_socket_path = journald_socket_path
        self.log_level_to_journald_priority = {
            logging.CRITICAL: 2,
            logging.ERROR: 3,
            logging.WARNING: 4,
            logging.ANSWER: 5,
            logging.INFO: 6,
            logging.DEBUG: 7,
        }

    def emit(self, record):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

        try:
            message_parts = []
            entry = dict(
                MESSAGE=record.getMessage(),
                PRIORITY=self.log_level_to_journald_priority.get(
                    record.levelno, DEFAULT_JOURNALD_PRIORITY
                ),
                SYSLOG_IDENTIFIER='borgmatic',
                SYSLOG_PID=os.getpid(),
            )

            for key, value in entry.items():
                encoded_key = key.upper().encode('utf-8')
                encoded_value = str(value).encode('utf-8')

                # Multi-line and single-line values use different formats on the wire.
                if b'\n' in encoded_value:
                    message_parts.extend((encoded_key, b'\n'))
                    message_parts.extend(
                        (len(encoded_value).to_bytes(8, 'little'), encoded_value, b'\n')
                    )
                else:
                    message_parts.extend((encoded_key, b'=', encoded_value, b'\n'))

            sock.sendto(b''.join(message_parts), self.journald_socket_path)
        finally:
            sock.close()


class Log_prefix_formatter(logging.Formatter):
    def __init__(self, fmt='{prefix}{message}', *args, style='{', **kwargs):
        self.prefix = None

        super().__init__(*args, fmt=fmt, style=style, **kwargs)

    def format(self, record):  # pragma: no cover
        record.prefix = f'{self.prefix}: ' if self.prefix else ''

        return super().format(record)


def log_record_to_json(record):
    '''
    Given a logging.LogRecord, return it as a JSON-encoded string containing relevant attributes.
    '''
    return json.dumps(
        dict(
            type='log_message',
            time=record.created,
            message=record.getMessage(),
            levelname=record.levelname,
            name=record.name,
        )
    )


class Json_formatter(logging.Formatter):
    def __init__(self, fmt='{message}', *args, style='{', **kwargs):
        super().__init__(*args, fmt=fmt, style=style, **kwargs)

    def format(self, record):  # noqa: PLR6301
        return log_record_to_json(record)


class Color(enum.Enum):
    RESET = 0
    RED = 31
    GREEN = 32
    YELLOW = 33
    MAGENTA = 35
    CYAN = 36


class Console_color_formatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        self.prefix = None
        super().__init__(
            '{prefix}{message}',
            *args,
            style='{',
            **kwargs,
        )

    def format(self, record):
        add_custom_log_levels()

        color = {
            logging.CRITICAL: Color.RED,
            logging.ERROR: Color.RED,
            logging.WARNING: Color.YELLOW,
            logging.ANSWER: Color.MAGENTA,
            logging.INFO: Color.GREEN,
            logging.DEBUG: Color.CYAN,
        }.get(record.levelno).value
        record.prefix = f'{self.prefix}: ' if self.prefix else ''

        return color_text(color, super().format(record))


def ansi_escape_code(color):  # pragma: no cover
    '''
    Given a color value, produce the corresponding ANSI escape code.
    '''
    return f'\x1b[{color}m'


def color_text(color, message):
    '''
    Given a color value and a message, return the message colored with ANSI escape codes.
    '''
    if not color:
        return message

    return f'{ansi_escape_code(color)}{message}{ansi_escape_code(Color.RESET.value)}'


def add_logging_level(level_name, level_number):
    '''
    Globally add a custom logging level based on the given (all uppercase) level name and number.
    Do this idempotently.

    Inspired by https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility/35804945#35804945
    '''
    method_name = level_name.lower()

    if not hasattr(logging, level_name):
        logging.addLevelName(level_number, level_name)
        setattr(logging, level_name, level_number)

    if not hasattr(logging, method_name):

        def log_for_level(self, message, *args, **kwargs):  # pragma: no cover
            if self.isEnabledFor(level_number):
                self._log(level_number, message, args, **kwargs)

        setattr(logging.getLoggerClass(), method_name, log_for_level)

    if not hasattr(logging.getLoggerClass(), method_name):

        def log_to_root(message, *args, **kwargs):  # pragma: no cover
            logging.log(level_number, message, *args, **kwargs)  # noqa: LOG015

        setattr(logging, method_name, log_to_root)


ANSWER = logging.WARNING - 5
DISABLED = logging.CRITICAL + 10


def add_custom_log_levels():  # pragma: no cover
    '''
    Add a custom log level between WARNING and INFO for user-requested answers.
    '''
    add_logging_level('ANSWER', ANSWER)
    add_logging_level('DISABLED', DISABLED)


def get_log_prefix():
    '''
    Return the current log prefix set by set_log_prefix(). Return None if no such prefix exists.

    It would be a whole lot easier to use logger.Formatter(defaults=...) instead, but that argument
    doesn't exist until Python 3.10+.
    '''
    try:
        formatter = next(
            handler.formatter
            for handler in logging.getLogger().handlers
            if handler.formatter
            if hasattr(handler.formatter, 'prefix')
        )
    except StopIteration:
        return None

    return formatter.prefix


def set_log_prefix(prefix):
    '''
    Given a log prefix as a string, set it into the each handler's formatter so that it can inject
    the prefix into each logged record.
    '''
    for handler in logging.getLogger().handlers:
        if handler.formatter and hasattr(handler.formatter, 'prefix'):
            handler.formatter.prefix = prefix


class Log_prefix:
    '''
    A Python context manager for setting a log prefix so that it shows up in every subsequent
    logging message for the duration of the context manager. For this to work, it relies on each
    logging formatter to be initialized with "{prefix}" somewhere in its logging format.

    Example use as a context manager:


       with borgmatic.logger.Log_prefix('myprefix'):
            do_something_that_logs()

    For the scope of that "with" statement, any logs created are prefixed with "myprefix: ".
    Afterwards, the prefix gets restored to whatever it was prior to the context manager.
    '''

    def __init__(self, prefix):
        '''
        Given the desired log prefix, save it for use below. Set prefix to None to disable any
        prefix from getting logged.
        '''
        self.prefix = prefix
        self.original_prefix = None

    def __enter__(self):
        '''
        Set the prefix onto the formatter defaults for every logging handler so that the prefix ends
        up in every log message. But first, save off any original prefix so that it can be restored
        below.
        '''
        self.original_prefix = get_log_prefix()
        set_log_prefix(self.prefix)

    def __exit__(self, exception_type, exception, traceback):
        '''
        Restore any original prefix.
        '''
        set_log_prefix(self.original_prefix)


class Delayed_logging_handler(logging.handlers.BufferingHandler):
    '''
    A logging handler that buffers logs and doesn't flush them until explicitly flushed (after
    target handlers are actually set). It's useful for holding onto messages logged before logging
    is configured, ensuring those records eventually make their way to the relevant logging
    handlers.

    When flushing, don't forward log records to a target handler if the record's log level is below
    that of the handler. This recreates the standard logging behavior of, say, logging.DEBUG records
    getting suppressed if a handler's level is only set to logging.INFO.
    '''

    def __init__(self):
        super().__init__(capacity=0)

        self.targets = None

    def shouldFlush(self, record):  # noqa: N802
        return self.targets is not None

    def flush(self):
        self.acquire()

        try:
            if not self.targets:
                return

            for record in self.buffer:
                for target in self.targets:
                    if record.levelno >= target.level:
                        target.handle(record)

            self.buffer.clear()
        finally:
            self.release()


def configure_delayed_logging():  # pragma: no cover
    '''
    Configure a delayed logging handler to buffer anything that gets logged until we're ready to
    deal with it.
    '''
    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[Delayed_logging_handler()],
    )


def flush_delayed_logging(target_handlers):
    '''
    Flush any previously buffered logs to our "real" logging handlers.
    '''
    root_logger = logging.getLogger()

    if root_logger.handlers and isinstance(root_logger.handlers[0], Delayed_logging_handler):
        delayed_handler = root_logger.handlers[0]
        delayed_handler.targets = target_handlers
        delayed_handler.flush()
        root_logger.removeHandler(delayed_handler)


JOURNALD_SOCKET_PATH = '/run/systemd/journal/socket'
SYSLOG_PATHS = ('/dev/log', '/var/run/syslog', '/var/run/log')


def configure_logging(
    console_log_level,
    syslog_log_level=None,
    log_file_log_level=None,
    monitoring_log_level=None,
    log_file=None,
    log_file_format=None,
    log_json=False,
    color_enabled=True,
):
    '''
    Configure logging to go to both the console and (syslog or log file). Use the given log levels,
    respectively. If color is enabled, set up log formatting accordingly.

    Raise FileNotFoundError or PermissionError if the log file could not be opened for writing.
    '''
    add_custom_log_levels()

    if syslog_log_level is None:
        syslog_log_level = logging.DISABLED

    if log_file_log_level is None:
        log_file_log_level = console_log_level

    if monitoring_log_level is None:
        monitoring_log_level = console_log_level

    # Log certain log levels to console stderr and others to stdout. This supports use cases like
    # grepping (non-error) output.
    console_disabled = logging.NullHandler()
    console_error_handler = logging.StreamHandler(sys.stderr)
    console_standard_handler = logging.StreamHandler(sys.stdout)
    console_handler = Multi_stream_handler(
        {
            logging.DISABLED: console_disabled,
            logging.CRITICAL: console_error_handler,
            logging.ERROR: console_error_handler,
            logging.WARNING: console_error_handler,
            logging.ANSWER: console_standard_handler,
            logging.INFO: console_standard_handler,
            logging.DEBUG: console_standard_handler,
        },
    )

    if log_json:
        console_handler.setFormatter(Json_formatter())
    elif color_enabled:
        console_handler.setFormatter(Console_color_formatter())
    else:
        console_handler.setFormatter(Log_prefix_formatter())

    console_handler.setLevel(console_log_level)
    handlers = [console_handler]

    if syslog_log_level != logging.DISABLED:
        if os.path.exists(JOURNALD_SOCKET_PATH):
            journald_handler = JournaldHandler(JOURNALD_SOCKET_PATH)
            journald_handler.setLevel(syslog_log_level)
            handlers.append(journald_handler)
        else:
            syslog_path = next(
                (path for path in SYSLOG_PATHS if os.path.exists(path)),
                None,
            )

            if syslog_path:
                syslog_handler = logging.handlers.SysLogHandler(address=syslog_path)
                syslog_handler.setFormatter(
                    Log_prefix_formatter(
                        'borgmatic: {levelname} {prefix}{message}',
                    ),
                )
                syslog_handler.setLevel(syslog_log_level)
                handlers.append(syslog_handler)

    if log_file and log_file_log_level != logging.DISABLED:
        file_handler = logging.handlers.WatchedFileHandler(log_file)
        file_handler.setFormatter(
            Json_formatter()
            if log_json
            else Log_prefix_formatter(
                log_file_format or '[{asctime}] {levelname}: {prefix}{message}',
            ),
        )
        file_handler.setLevel(log_file_log_level)
        handlers.append(file_handler)

    flush_delayed_logging(handlers)

    logging.basicConfig(
        level=min(handler.level for handler in handlers),
        handlers=handlers,
    )



import fnmatch
import os

import jsonschema
import ruamel.yaml

import borgmatic.config.arguments
from borgmatic.config import constants, environment, load, normalize, override


def schema_filename():
    '''
    Path to the installed YAML configuration schema file, used to validate and parse the
    configuration.

    Raise FileNotFoundError when the schema path does not exist.
    '''
    schema_path = os.path.join(os.path.dirname(borgmatic.config.__file__), 'schema.yaml')

    with open(schema_path, encoding='utf-8'):
        return schema_path


def load_schema(schema_path):  # pragma: no cover
    '''
    Given a schema filename path, load the schema and return it as a dict.

    Raise Validation_error if the schema could not be parsed.
    '''
    try:
        return load.load_configuration(schema_path)
    except (ruamel.yaml.error.YAMLError, RecursionError) as error:
        raise Validation_error(schema_path, (str(error),))


def format_json_error_path_element(path_element):
    '''
    Given a path element into a JSON data structure, format it for display as a string.
    '''
    if isinstance(path_element, int):
        return str(f'[{path_element}]')

    return str(f'.{path_element}')


def format_json_error(error):
    '''
    Given an instance of jsonschema.exceptions.ValidationError, format it for display as a string.
    '''
    if not error.path:
        return f'At the top level: {error.message}'

    formatted_path = ''.join(format_json_error_path_element(element) for element in error.path)
    return f"At '{formatted_path.lstrip('.')}': {error.message}"


class Validation_error(ValueError):
    '''
    A collection of error messages generated when attempting to validate a particular
    configuration file.
    '''

    def __init__(self, config_filename, errors):
        '''
        Given a configuration filename path and a sequence of string error messages, create a
        Validation_error.
        '''
        self.config_filename = config_filename
        self.errors = errors

    def __str__(self):
        '''
        Render a validation error as a user-facing string.
        '''
        return (
            f'An error occurred while parsing a configuration file at {self.config_filename}:\n'
            + '\n'.join(error for error in self.errors)
        )


def apply_logical_validation(config_filename, parsed_configuration):
    '''
    Given a parsed and schematically valid configuration as a data structure of nested dicts (see
    below), run through any additional logical validation checks. If there are any such validation
    problems, raise a Validation_error.
    '''
    repositories = parsed_configuration.get('repositories')
    check_repositories = parsed_configuration.get('check_repositories', [])
    for repository in check_repositories:
        if not any(
            repositories_match(repository, config_repository) for config_repository in repositories
        ):
            raise Validation_error(
                config_filename,
                (f'Unknown repository in "check_repositories": {repository}',),
            )


def parse_configuration(
    config_filename,
    schema_filename,
    arguments,
    overrides=None,
    resolve_env=True,
):
    '''
    Given the path to a config filename in YAML format, the path to a schema filename in a YAML
    rendition of JSON Schema format, arguments as dict from action name to argparse.Namespace, a
    sequence of configuration file override strings in the form of "option.suboption=value", and
    whether to resolve environment variables, return the parsed configuration as a data structure of
    nested dicts and lists corresponding to the schema. Example return value.

    Example return value:

        {
            'source_directories': ['/home', '/etc'],
            'repository': 'hostname.borg',
            'keep_daily': 7,
            'checks': ['repository', 'archives'],
        }

    Also return a set of loaded configuration paths and a sequence of logging.LogRecord instances
    containing any warnings about the configuration.

    Raise FileNotFoundError if the file does not exist, PermissionError if the user does not
    have permissions to read the file, or Validation_error if the config does not match the schema.
    '''
    config_paths = set()

    try:
        config = load.load_configuration(config_filename, config_paths)
        schema = load.load_configuration(schema_filename)
    except (ruamel.yaml.error.YAMLError, RecursionError) as error:
        raise Validation_error(config_filename, (str(error),))

    borgmatic.config.arguments.apply_arguments_to_config(config, schema, arguments)
    override.apply_overrides(config, schema, overrides)
    constants.apply_constants(config, config.get('constants') if config else {})

    if resolve_env:
        environment.resolve_env_variables(config)

    logs = normalize.normalize(config_filename, config)

    try:
        validator = jsonschema.Draft7Validator(schema)
    except AttributeError:  # pragma: no cover
        validator = jsonschema.Draft4Validator(schema)

    validation_errors = tuple(validator.iter_errors(config))

    if validation_errors:
        raise Validation_error(
            config_filename,
            tuple(format_json_error(error) for error in validation_errors),
        )

    apply_logical_validation(config_filename, config)

    return config, config_paths, logs


def normalize_repository_path(repository, base=None):
    '''
    Given a repository path, return the absolute path of it (for local repositories).
    Optionally, use a base path for resolving relative paths, e.g. to the configured working directory.
    '''
    # A colon in the repository could mean that it's either a file:// URL or a remote repository.
    # If it's a remote repository, we don't want to normalize it. If it's a file:// URL, we do.
    if ':' not in repository:
        return (
            os.path.abspath(os.path.join(base, repository)) if base else os.path.abspath(repository)
        )

    if repository.startswith('file://'):
        local_path = repository.partition('file://')[-1]
        return (
            os.path.abspath(os.path.join(base, local_path)) if base else os.path.abspath(local_path)
        )

    return repository


def glob_match(first, second):
    '''
    Given two strings, return whether the first matches the second. Globs are
    supported.
    '''
    if first is None or second is None:
        return False

    return fnmatch.fnmatch(first, second) or fnmatch.fnmatch(second, first)


def repositories_match(first, second):
    '''
    Given two repository dicts with keys "path" (relative and/or absolute),
    and "label", two repository paths as strings, or a mix of the two formats,
    return whether they match. Globs are supported.
    '''
    if isinstance(first, str):
        first = {'path': first, 'label': first}

    if isinstance(second, str):
        second = {'path': second, 'label': second}

    return glob_match(first.get('label'), second.get('label')) or glob_match(
        normalize_repository_path(first.get('path')),
        normalize_repository_path(second.get('path')),
    )


def guard_configuration_contains_repository(repository, configurations):
    '''
    Given a repository path and a dict mapping from config filename to corresponding parsed config
    dict, ensure that the repository is declared at least once in all of the configurations. If no
    repository is given, skip this check.

    Raise ValueError if the repository is not found in any configurations.
    '''
    if not repository:
        return

    count = len(
        tuple(
            config_repository
            for config in configurations.values()
            for config_repository in config['repositories']
            if repositories_match(config_repository, repository)
        ),
    )

    if count == 0:
        raise ValueError(f'Repository "{repository}" not found in configuration files')



import functools
import re
import shlex

import borgmatic.hooks.dispatch

IS_A_HOOK = False


class Hash_adapter:
    '''
    A Hash_adapter instance wraps an unhashable object and pretends it's hashable. This is intended
    for passing to a @functools.cache-decorated function to prevent it from complaining that an
    argument is unhashable. It should only be used for arguments that you don't want to actually
    impact the cache hashing, because Hash_adapter doesn't actually hash the object's contents.

    Example usage:

        @functools.cache
        def func(a, b):
            print(a, b.actual_value)
            return a

        func(5, Hash_adapter({1: 2, 3: 4}))  # Calls func(), prints, and returns.
        func(5, Hash_adapter({1: 2, 3: 4}))  # Hits the cache and just returns the value.
        func(5, Hash_adapter({5: 6, 7: 8}))  # Also uses cache, since the Hash_adapter is ignored.

    In the above function, the "b" value is one that has been wrapped with Hash_adappter, and
    therefore "b.actual_value" is necessary to access the original value.
    '''

    def __init__(self, actual_value):
        self.actual_value = actual_value

    def __eq__(self, other):
        return True

    def __hash__(self):
        return 0


UNHASHABLE_TYPES = (dict, list, set)


def cache_ignoring_unhashable_arguments(function):
    '''
    A function decorator that caches calls to the decorated function but ignores any unhashable
    arguments when performing cache lookups. This is intended to be a drop-in replacement for
    functools.cache.

    Example usage:

        @cache_ignoring_unhashable_arguments
        def func(a, b):
            print(a, b)
            return a

        func(5, {1: 2, 3: 4})  # Calls func(), prints, and returns.
        func(5, {1: 2, 3: 4})  # Hits the cache and just returns the value.
        func(5, {5: 6, 7: 8})  # Also uses cache, since the unhashable value (the dict) is ignored.
    '''

    @functools.cache
    def cached_function(*args, **kwargs):
        return function(
            *(arg.actual_value if isinstance(arg, Hash_adapter) else arg for arg in args),
            **{
                key: value.actual_value if isinstance(value, Hash_adapter) else value
                for (key, value) in kwargs.items()
            },
        )

    @functools.wraps(function)
    def wrapper_function(*args, **kwargs):
        return cached_function(
            *(Hash_adapter(arg) if isinstance(arg, UNHASHABLE_TYPES) else arg for arg in args),
            **{
                key: Hash_adapter(value) if isinstance(value, UNHASHABLE_TYPES) else value
                for (key, value) in kwargs.items()
            },
        )

    wrapper_function.cache_clear = cached_function.cache_clear

    return wrapper_function


CREDENTIAL_PATTERN = re.compile(r'\{credential( +(?P<hook_and_parameters>.*))?\}')


@cache_ignoring_unhashable_arguments
def resolve_credential(value, config):
    '''
    Given a configuration value containing a string like "{credential hookname credentialname}" and
    a configuration dict, resolve the credential by calling the relevant hook to get the actual
    credential value. If the given value does not actually contain a credential tag, then return it
    unchanged.

    Cache the value (ignoring the config for purposes of caching), so repeated calls to this
    function don't need to load the credential repeatedly.

    Raise ValueError if the config could not be parsed or the credential could not be loaded.
    '''
    if value is None:
        return value

    matcher = CREDENTIAL_PATTERN.match(value)

    if not matcher:
        return value

    hook_and_parameters = matcher.group('hook_and_parameters')

    if not hook_and_parameters:
        raise ValueError(f'Cannot load credential with invalid syntax "{value}"')

    (hook_name, *credential_parameters) = shlex.split(hook_and_parameters)

    if not credential_parameters:
        raise ValueError(f'Cannot load credential with invalid syntax "{value}"')

    return borgmatic.hooks.dispatch.call_hook(
        'load_credential',
        config,
        hook_name,
        tuple(credential_parameters),
    )



import logging

import requests

from borgmatic.hooks.monitoring import monitor

logger = logging.getLogger(__name__)

MONITOR_STATE_TO_CRONHUB = {
    monitor.State.START: 'start',
    monitor.State.FINISH: 'finish',
    monitor.State.FAIL: 'fail',
}
TIMEOUT_SECONDS = 10


def initialize_monitor(
    ping_url,
    config,
    config_filename,
    monitoring_log_level,
    dry_run,
):  # pragma: no cover
    '''
    No initialization is necessary for this monitor.
    '''


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    Ping the configured Cronhub URL, modified with the monitor.State. Use the given configuration
    filename in any log entries. If this is a dry run, then don't actually ping anything.
    '''
    if state not in MONITOR_STATE_TO_CRONHUB:
        logger.debug(f'Ignoring unsupported monitoring state {state.name.lower()} in Cronhub hook')
        return

    dry_run_label = ' (dry run; not actually pinging)' if dry_run else ''
    formatted_state = f'/{MONITOR_STATE_TO_CRONHUB[state]}/'
    ping_url = (
        hook_config['ping_url']
        .replace('/start/', formatted_state)
        .replace('/ping/', formatted_state)
    )

    logger.info(f'Pinging Cronhub {state.name.lower()}{dry_run_label}')
    logger.debug(f'Using Cronhub ping URL {ping_url}')

    if not dry_run:
        logging.getLogger('urllib3').setLevel(logging.ERROR)
        try:
            response = requests.get(
                ping_url,
                timeout=TIMEOUT_SECONDS,
                headers={'User-Agent': 'borgmatic'},
            )
            if not response.ok:
                response.raise_for_status()
        except requests.exceptions.RequestException as error:
            logger.warning(f'Cronhub error: {error}')


def destroy_monitor(ping_url_or_uuid, config, monitoring_log_level, dry_run):  # pragma: no cover
    '''
    No destruction is necessary for this monitor.
    '''



import logging

import borgmatic.borg.mount
import borgmatic.borg.repo_list

logger = logging.getLogger(__name__)


def run_mount(
    repository,
    config,
    local_borg_version,
    mount_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "mount" action for the given repository.
    '''
    if mount_arguments.archive:
        logger.info(f'Mounting archive {mount_arguments.archive}')
    else:  # pragma: nocover
        logger.info('Mounting repository')

    borgmatic.borg.mount.mount_archive(
        repository['path'],
        borgmatic.borg.repo_list.resolve_archive_name(
            repository['path'],
            mount_arguments.archive,
            config,
            local_borg_version,
            global_arguments,
            local_path,
            remote_path,
        ),
        mount_arguments,
        config,
        local_borg_version,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )



import enum
import importlib
import logging
import pkgutil

import borgmatic.hooks.credential
import borgmatic.hooks.data_source
import borgmatic.hooks.monitoring

logger = logging.getLogger(__name__)


class Hook_type(enum.Enum):
    CREDENTIAL = 'credential'
    DATA_SOURCE = 'data_source'
    MONITORING = 'monitoring'


def get_submodule_names(parent_module):  # pragma: no cover
    '''
    Given a parent module, return the names of its direct submodules as a tuple of strings.
    '''
    return tuple(module_info.name for module_info in pkgutil.iter_modules(parent_module.__path__))


def call_hook(function_name, config, hook_name, *args, **kwargs):
    '''
    Given a configuration dict, call the requested function of the Python module corresponding to
    the given hook name. Supply that call with the configuration for this hook (if any) and any
    given args and kwargs. Return the return value of that call or None if the module in question is
    not a hook.

    Raise ValueError if the hook name is unknown.
    Raise AttributeError if the function name is not found in the module.
    Raise anything else that the called function raises.
    '''
    if hook_name in config or f'{hook_name}_databases' in config:
        hook_config = config.get(hook_name) or config.get(f'{hook_name}_databases') or {}
    else:
        hook_config = None

    module_name = hook_name.split('_databases')[0]

    # Probe for a data source or monitoring hook module corresponding to the hook name.
    for parent_module in (
        borgmatic.hooks.credential,
        borgmatic.hooks.data_source,
        borgmatic.hooks.monitoring,
    ):
        if module_name not in get_submodule_names(parent_module):
            continue

        module = importlib.import_module(f'{parent_module.__name__}.{module_name}')

        # If this module is explicitly flagged as not a hook, bail.
        if not getattr(module, 'IS_A_HOOK', True):
            return None

        break
    else:
        raise ValueError(f'Unknown hook name: {hook_name}')

    logger.debug(f'Calling {hook_name} hook function {function_name}')

    return getattr(module, function_name)(hook_config, config, *args, **kwargs)


def call_hooks(function_name, config, hook_type, *args, **kwargs):
    '''
    Given a configuration dict, call the requested function of the Python module corresponding to
    each hook of the given hook type ("credential", "data_source", or "monitoring"). Supply each
    call with the configuration for that hook, and any given args and kwargs.

    Collect any return values into a dict from module name to return value. Note that the module
    name is the name of the hook module itself, which might be different from the hook configuration
    option (e.g. "postgresql" for the former vs. "postgresql_databases" for the latter).

    If the hook name is not present in the hooks configuration, then don't call the function for it
    and omit it from the return values.

    Raise ValueError if the hook name is unknown.
    Raise AttributeError if the function name is not found in the module.
    Raise anything else that a called function raises. An error stops calls to subsequent functions.
    '''
    return {
        hook_name: call_hook(function_name, config, hook_name, *args, **kwargs)
        for hook_name in get_submodule_names(
            importlib.import_module(f'borgmatic.hooks.{hook_type.value}'),
        )
        if hook_name in config or f'{hook_name}_databases' in config
    }


def call_hooks_even_if_unconfigured(function_name, config, hook_type, *args, **kwargs):
    '''
    Given a configuration dict, call the requested function of the Python module corresponding to
    each hook of the given hook type ("credential", "data_source", or "monitoring"). Supply each
    call with the configuration for that hook and any given args and kwargs. Collect any return
    values into a dict from hook name to return value.

    Raise AttributeError if the function name is not found in the module.
    Raise anything else that a called function raises. An error stops calls to subsequent functions.
    '''
    return {
        hook_name: call_hook(function_name, config, hook_name, *args, **kwargs)
        for hook_name in get_submodule_names(
            importlib.import_module(f'borgmatic.hooks.{hook_type.value}'),
        )
    }



import logging
import shlex

import borgmatic.config.paths
from borgmatic.execute import execute_command

logger = logging.getLogger(__name__)


def unmount_archive(config, mount_point, local_path='borg'):
    '''
    Given a mounted filesystem mount point, and an optional local Borg paths, umount the filesystem
    from the mount point.
    '''
    extra_borg_options = config.get('extra_borg_options', {}).get('umount', '')
    full_command = (
        (local_path, 'umount')
        + ('--log-json',)
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + (mount_point,)
    )

    execute_command(
        full_command,
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import glob
import itertools
import logging
import os
import pathlib

import borgmatic.borg.pattern

logger = logging.getLogger(__name__)


def parse_pattern(pattern_line, default_style=borgmatic.borg.pattern.Pattern_style.NONE):
    '''
    Given a Borg pattern as a string, parse it into a borgmatic.borg.pattern.Pattern instance and
    return it.
    '''
    try:
        (pattern_type, remainder) = pattern_line.split(' ', maxsplit=1)
    except ValueError:
        raise ValueError(f'Invalid pattern: {pattern_line}')

    try:
        (parsed_pattern_style, path) = remainder.split(':', maxsplit=1)
        pattern_style = borgmatic.borg.pattern.Pattern_style(parsed_pattern_style)
    except ValueError:
        pattern_style = default_style
        path = remainder

    return borgmatic.borg.pattern.Pattern(
        path,
        borgmatic.borg.pattern.Pattern_type(pattern_type),
        borgmatic.borg.pattern.Pattern_style(pattern_style),
        source=borgmatic.borg.pattern.Pattern_source.CONFIG,
    )


def collect_patterns(config):
    '''
    Given a configuration dict, produce a single sequence of patterns comprised of the configured
    source directories, patterns, excludes, pattern files, and exclude files.

    The idea is that Borg has all these different ways of specifying includes, excludes, source
    directories, etc., but we'd like to collapse them all down to one common format (patterns) for
    ease of manipulation within borgmatic.
    '''
    try:
        return (
            tuple(
                borgmatic.borg.pattern.Pattern(
                    source_directory,
                    source=borgmatic.borg.pattern.Pattern_source.CONFIG,
                )
                for source_directory in config.get('source_directories', ())
            )
            + tuple(
                parse_pattern(pattern_line.strip())
                for pattern_line in config.get('patterns', ())
                if not pattern_line.lstrip().startswith('#')
                if pattern_line.strip()
            )
            + tuple(
                parse_pattern(
                    f'{borgmatic.borg.pattern.Pattern_type.NO_RECURSE.value} {exclude_line.strip()}',
                    borgmatic.borg.pattern.Pattern_style.FNMATCH,
                )
                for exclude_line in config.get('exclude_patterns', ())
            )
            + tuple(
                parse_pattern(pattern_line.strip())
                for filename in config.get('patterns_from', ())
                for pattern_line in open(filename, encoding='utf-8')
                if not pattern_line.lstrip().startswith('#')
                if pattern_line.strip()
            )
            + tuple(
                parse_pattern(
                    f'{borgmatic.borg.pattern.Pattern_type.NO_RECURSE.value} {exclude_line.strip()}',
                    borgmatic.borg.pattern.Pattern_style.FNMATCH,
                )
                for filename in config.get('exclude_from', ())
                for exclude_line in open(filename, encoding='utf-8')
                if not exclude_line.lstrip().startswith('#')
                if exclude_line.strip()
            )
        )
    except (FileNotFoundError, OSError) as error:
        logger.debug(error)

        raise ValueError(f'Cannot read patterns_from/exclude_from file: {error.filename}')


def expand_directory(directory, working_directory):
    '''
    Given a directory path, expand any tilde (representing a user's home directory) and any globs
    therein. Return a list of one or more resulting paths.

    Take into account the given working directory so that relative paths are supported.
    '''
    expanded_directory = os.path.expanduser(directory)

    # This would be a lot easier to do with glob(..., root_dir=working_directory), but root_dir is
    # only available in Python 3.10+.
    normalized_directory = os.path.join(working_directory or '', expanded_directory)
    glob_paths = glob.glob(normalized_directory)

    if not glob_paths:
        return [expanded_directory]

    working_directory_prefix = os.path.join(working_directory or '', '')

    return [
        (
            glob_path
            # If these are equal, that means we didn't add any working directory prefix above.
            if normalized_directory == expanded_directory
            # Remove the working directory prefix added above in order to make glob() work. We
            # can't use os.path.relpath() here because it collapses any use of Borg's slashdot hack.
            else glob_path.removeprefix(working_directory_prefix)
        )
        for glob_path in glob_paths
    ]


def expand_patterns(patterns, working_directory=None, skip_paths=None):
    '''
    Given a sequence of borgmatic.borg.pattern.Pattern instances and an optional working directory,
    expand tildes and globs in each root pattern and expand just tildes in each non-root pattern.
    The idea is that non-root patterns may be regular expressions or other pattern styles containing
    "*" that borgmatic should not expand as a shell glob.

    Return all the resulting patterns as a tuple.

    If a set of paths are given to skip, then don't expand any patterns matching them.
    '''
    if patterns is None:
        return ()

    return tuple(
        itertools.chain.from_iterable(
            (
                (
                    borgmatic.borg.pattern.Pattern(
                        expanded_path,
                        pattern.type,
                        pattern.style,
                        pattern.device,
                        pattern.source,
                    )
                    for expanded_path in expand_directory(pattern.path, working_directory)
                )
                if pattern.type == borgmatic.borg.pattern.Pattern_type.ROOT
                and pattern.path not in (skip_paths or ())
                else (
                    borgmatic.borg.pattern.Pattern(
                        os.path.expanduser(pattern.path),
                        pattern.type,
                        pattern.style,
                        pattern.device,
                        pattern.source,
                    ),
                )
            )
            for pattern in patterns
        ),
    )


def get_existent_path_or_parent(path):
    '''
    Given a path, return it if it exists. Otherwise, return the longest parent directory of the path
    that exists. Return None if none of these paths exist.

    This is used below for finding an existent path prefix of pattern's path, which is necessary if
    the path contain globs or other special characters that we don't want to try to interpret
    (because we want to leave that responsibility to Borg).
    '''
    if path.startswith('/e2e/'):
        return None

    try:
        return next(
            candidate_path
            for candidate_path in (
                path,
                *tuple(str(parent) for parent in pathlib.PurePath(path).parents),
            )
            if os.path.exists(candidate_path)
        )
    except StopIteration:
        return None


def device_map_patterns(patterns, working_directory=None):
    '''
    Given a sequence of borgmatic.borg.pattern.Pattern instances and an optional working directory,
    determine the identifier for the device on which the pattern's path resides—or None if the path
    doesn't exist or is from a non-root pattern. Return an updated sequence of patterns with the
    device field populated. But if the device field is already set, don't bother setting it again.

    This is handy for determining whether two different pattern paths are on the same filesystem
    (have the same device identifier).

    This function only considers the start of a pattern's path—from the start of the path up until
    there's a path component with a glob or other non-literal character. If there are no such
    characters, the whole path is considered. The rationale is that it's not feasible for borgmatic
    to interpret Borg's patterns to see which actual files (and therefore devices) they map to. So
    for instance, a pattern with a path of "/var/log/*/data" would end up with its device set to the
    device of "/var/log"—ignoring the "/*/data" part due to that glob.

    The one exception is that if a regular expression pattern path starts with "^", that will get
    stripped off for purposes of determining its device.
    '''
    return tuple(
        borgmatic.borg.pattern.Pattern(
            pattern.path,
            pattern.type,
            pattern.style,
            device=pattern.device or (os.stat(existent_path).st_dev if existent_path else None),
            source=pattern.source,
        )
        for pattern in patterns
        for existent_path in (
            get_existent_path_or_parent(
                os.path.join(working_directory or '', pattern.path.lstrip('^')),
            ),
        )
    )


def deduplicate_runtime_directory_patterns(patterns, config, borgmatic_runtime_directory=None):
    '''
    Given a sequence of borgmatic.borg.pattern.Pattern instances, the borgmatic runtime directory,
    and a configuration dict, return them without any duplicate root child patterns that contain the
    runtime directory. For instance, if two root patterns are given with paths "/foo" and
    "/foo/bar", and the runtime directory is "/foo/bar", return just the "/foo" pattern. Non-root
    patterns and patterns not containing the runtime directory are passed through without
    modification.

    One exception to deduplication is if two paths are on different filesystems (devices) and
    "one_file_system" is True in the given configuration. In that case, the paths won't get
    deduplicated, because Borg won't cross filesystem boundaries when "one_file_system" is True.

    The idea is that if Borg is given a root parent pattern containing the borgmatic runtime
    directory, then Borg doesn't also need to be given child patterns, because it will naturally
    spider the contents of the parent pattern's path. Additionally, there are cases where Borg
    coming across the same file twice will result in duplicate reads and even hangs, e.g. when a
    database hook in the borgmatic runtime directory is using a named pipe for streaming database
    dumps to Borg.

    This deduplication is limited to the borgmatic runtime directory (where borgmatic's named pipes
    exist), because there are other legitimate use cases for parent and child patterns to both exist
    in patterns. For instance, with some snapshotted filesystems, snapshots don't traverse from a
    parent filesystem to a child and therefore both need to remain in patterns.

    And for the case of named pipes outside of the borgmatic runtime directory, there is code
    elsewhere (in the "create" action) that auto-excludes special files to prevent Borg hangs.
    '''
    if borgmatic_runtime_directory is None:
        return patterns

    deduplicated = {}  # Use just the keys as an ordered set.
    runtime_directory_parents = set(pathlib.PurePath(borgmatic_runtime_directory).parents).union(
        {pathlib.PurePath(borgmatic_runtime_directory)}
    )

    for pattern in patterns:
        if pattern.type != borgmatic.borg.pattern.Pattern_type.ROOT:
            deduplicated[pattern] = True
            continue

        pattern_parents = pathlib.PurePath(pattern.path).parents

        # If:
        #
        #   1. another pattern is a parent of the current pattern (even n levels up),
        #   2. both patterns are parents of the runtime directory (even n levels up),
        #   3. and both patterns are on the same filesystem (or one_file_system is not set)
        #
        # ... then consider the current pattern as a duplicate.
        for other_pattern in patterns:
            if other_pattern.type != borgmatic.borg.pattern.Pattern_type.ROOT:
                continue

            device_matches = pattern.device is not None and (
                other_pattern.device == pattern.device or config.get('one_file_system') is not True
            )

            if (
                pathlib.PurePath(other_pattern.path) in pattern_parents
                and pathlib.PurePosixPath(other_pattern.path) in runtime_directory_parents
                and pathlib.PurePosixPath(pattern.path) in runtime_directory_parents
                and device_matches
            ):
                break
        else:
            deduplicated[pattern] = True

    return tuple(deduplicated.keys())


def process_patterns(
    patterns, config, working_directory, borgmatic_runtime_directory=None, skip_expand_paths=None
):
    '''
    Given a sequence of Borg patterns, a configuration dict, a configured working directory, the
    borgmatic runtime directory, and a sequence of paths to skip path expansion for, expand and
    deduplicate any "root" patterns, returning the resulting root and non-root patterns as a list.

    If the borgmatic runtime directory is None, then don't deduplicate patterns. Deduplication is
    really only necessary for the "create" action when the runtime directory might contain named
    pipes for database dumps.
    '''
    skip_paths = set(skip_expand_paths or ())

    return list(
        deduplicate_runtime_directory_patterns(
            device_map_patterns(
                expand_patterns(
                    patterns,
                    working_directory=working_directory,
                    skip_paths=skip_paths,
                ),
            ),
            config,
            borgmatic_runtime_directory,
        ),
    )



import collections
import importlib.metadata
import json
import logging
import os
import sys
import time
from queue import Queue
from subprocess import CalledProcessError

import ruamel.yaml

import borgmatic.actions.borg
import borgmatic.actions.break_lock
import borgmatic.actions.change_passphrase
import borgmatic.actions.check
import borgmatic.actions.compact
import borgmatic.actions.config.bootstrap
import borgmatic.actions.config.generate
import borgmatic.actions.config.show
import borgmatic.actions.config.validate
import borgmatic.actions.create
import borgmatic.actions.delete
import borgmatic.actions.diff
import borgmatic.actions.export_key
import borgmatic.actions.export_tar
import borgmatic.actions.extract
import borgmatic.actions.import_key
import borgmatic.actions.info
import borgmatic.actions.list
import borgmatic.actions.mount
import borgmatic.actions.prune
import borgmatic.actions.recreate
import borgmatic.actions.repo_create
import borgmatic.actions.repo_delete
import borgmatic.actions.repo_info
import borgmatic.actions.repo_list
import borgmatic.actions.restore
import borgmatic.actions.transfer
import borgmatic.commands.completion.bash
import borgmatic.commands.completion.fish
import borgmatic.config.load
import borgmatic.config.paths
from borgmatic.borg import umount as borg_umount
from borgmatic.borg import version as borg_version
from borgmatic.commands.arguments import parse_arguments
from borgmatic.config import checks, collect, validate
from borgmatic.hooks import command, dispatch
from borgmatic.hooks.monitoring import monitor
from borgmatic.logger import (
    DISABLED,
    Log_prefix,
    add_custom_log_levels,
    configure_delayed_logging,
    configure_logging,
    should_do_markup,
)
from borgmatic.signals import configure_signals
from borgmatic.verbosity import get_verbosity, verbosity_to_log_level

logger = logging.getLogger(__name__)


def get_skip_actions(config, arguments):
    '''
    Given a configuration dict and command-line arguments as an argparse.Namespace, return a list of
    the configured action names to skip. Omit "check" from this list though if "check --force" is
    part of the command-like arguments.
    '''
    skip_actions = config.get('skip_actions', [])

    if 'check' in arguments and arguments['check'].force:
        return [action for action in skip_actions if action != 'check']

    return skip_actions


class Monitoring_hooks:
    '''
    A Python context manager for pinging monitoring hooks for the start state before the wrapped
    code and log and finish (or failure) after the wrapped code. Also responsible for
    initializing/destroying the monitoring hooks.

    Example use as a context manager:

       with Monitoring_hooks(config_filename, config, arguments, global_arguments):
           do_stuff()
    '''

    def __init__(self, config_filename, config, arguments, global_arguments):
        '''
        Given a configuration filename, a configuration dict, command-line arguments as an
        argparse.Namespace, and global arguments as an argparse.Namespace, save relevant data points
        for use below.
        '''
        using_primary_action = {'create', 'prune', 'compact', 'check'}.intersection(arguments)
        self.config_filename = config_filename
        self.config = config
        self.dry_run = global_arguments.dry_run
        self.monitoring_log_level = verbosity_to_log_level(
            get_verbosity({config_filename: config}, 'monitoring_verbosity'),
        )
        self.monitoring_hooks_are_activated = (
            using_primary_action and self.monitoring_log_level != DISABLED
        )

    def __enter__(self):
        '''
        If monitoring hooks are enabled and a primary action is in use, initialize monitoring hooks
        and ping them for the "start" state.
        '''
        if not self.monitoring_hooks_are_activated:
            return

        dispatch.call_hooks(
            'initialize_monitor',
            self.config,
            dispatch.Hook_type.MONITORING,
            self.config_filename,
            self.monitoring_log_level,
            self.dry_run,
        )

        try:
            dispatch.call_hooks(
                'ping_monitor',
                self.config,
                dispatch.Hook_type.MONITORING,
                self.config_filename,
                monitor.State.START,
                self.monitoring_log_level,
                self.dry_run,
            )
        except (OSError, CalledProcessError) as error:
            raise ValueError(f'Error pinging monitor: {error}')

    def __exit__(self, exception_type, exception, traceback):
        '''
        If monitoring hooks are enabled and a primary action is in use, ping monitoring hooks for
        the "log" state and also the "finish" or "fail" states (depending on whether there's an
        exception). Lastly, destroy monitoring hooks.
        '''
        if not self.monitoring_hooks_are_activated:
            return

        # Send logs irrespective of error.
        try:
            dispatch.call_hooks(
                'ping_monitor',
                self.config,
                dispatch.Hook_type.MONITORING,
                self.config_filename,
                monitor.State.LOG,
                self.monitoring_log_level,
                self.dry_run,
            )
        except (OSError, CalledProcessError) as error:
            raise ValueError(f'Error pinging monitor: {error}')

        try:
            dispatch.call_hooks(
                'ping_monitor',
                self.config,
                dispatch.Hook_type.MONITORING,
                self.config_filename,
                monitor.State.FAIL if exception else monitor.State.FINISH,
                self.monitoring_log_level,
                self.dry_run,
            )
        except (OSError, CalledProcessError) as error:
            # If the wrapped code errored, prefer raising that exception, as it's probably more
            # important than a monitor failing to ping.
            if exception:
                return

            raise ValueError(f'Error pinging monitor: {error}')

        dispatch.call_hooks(
            'destroy_monitor',
            self.config,
            dispatch.Hook_type.MONITORING,
            self.monitoring_log_level,
            self.dry_run,
        )


def run_configuration(config_filename, config, config_paths, arguments):  # noqa: PLR0912, PLR0915
    '''
    Given a config filename, the corresponding parsed config dict, a sequence of loaded
    configuration paths, and command-line arguments as a dict from subparser name to a namespace of
    parsed arguments, execute the defined create, prune, compact, check, and/or other actions.

    Yield a combination of:

      * JSON output strings from successfully executing any actions that produce JSON
      * logging.LogRecord instances containing errors from any actions or backup hooks that fail
    '''
    global_arguments = arguments['global']

    local_path = config.get('local_path', 'borg')
    remote_path = config.get('remote_path')
    retries = config.get('retries', 0)
    retry_wait = config.get('retry_wait', 0)
    repo_queue = Queue()
    encountered_error = None
    error_repository = None
    skip_actions = get_skip_actions(config, arguments)

    if skip_actions:
        logger.debug(
            f"Skipping {'/'.join(skip_actions)} action{'s' if len(skip_actions) > 1 else ''} due to configured skip_actions",
        )

    try:  # noqa: PLR1702
        with (
            Monitoring_hooks(config_filename, config, arguments, global_arguments),
            borgmatic.hooks.command.Before_after_hooks(
                command_hooks=config.get('commands'),
                before_after='configuration',
                umask=config.get('umask'),
                working_directory=borgmatic.config.paths.get_working_directory(config),
                dry_run=global_arguments.dry_run,
                action_names=arguments.keys(),
                configuration_filename=config_filename,
                log_file=config.get('log_file', ''),
            ),
        ):
            try:
                local_borg_version = borg_version.local_borg_version(config, local_path)
                logger.debug(f'Borg {local_borg_version}')
            except (OSError, CalledProcessError, ValueError) as error:
                yield from log_error_records(
                    f'{config_filename}: Error getting local Borg version',
                    error,
                )
                raise

            for repo in config['repositories']:
                repo_queue.put(
                    (repo, 0),
                )

            while not repo_queue.empty():
                repository, retry_num = repo_queue.get()

                with Log_prefix(repository.get('label', repository['path'])):
                    logger.debug('Running actions for repository')
                    timeout = retry_num * retry_wait

                    if timeout:
                        logger.warning(f'Sleeping {timeout}s before next retry')
                        time.sleep(timeout)

                    try:
                        yield from run_actions(
                            arguments=arguments,
                            config_filename=config_filename,
                            config=config,
                            config_paths=config_paths,
                            local_path=local_path,
                            remote_path=remote_path,
                            local_borg_version=local_borg_version,
                            repository=repository,
                        )
                    except (OSError, CalledProcessError, ValueError) as error:
                        if retry_num < retries:
                            repo_queue.put(
                                (repository, retry_num + 1),
                            )
                            tuple(  # Consume the generator so as to trigger logging.
                                log_error_records(
                                    'Error running actions for repository',
                                    error,
                                    levelno=logging.WARNING,
                                    log_command_error_output=True,
                                ),
                            )
                            logger.warning(f'Retrying... attempt {retry_num + 1}/{retries}')
                            continue

                        if command.considered_soft_failure(error):
                            continue

                        yield from log_error_records(
                            'Error running actions for repository',
                            error,
                        )
                        encountered_error = error
                        error_repository = repository

            # Re-raise any error, so that the Monitoring_hooks context manager wrapping this
            # code can see the error and act accordingly. Do this here rather than as soon as
            # the error is encountered so that an error with one repository doesn't prevent
            # other repositories from running.
            if encountered_error:
                raise encountered_error

    except (OSError, CalledProcessError, ValueError) as error:
        yield from log_error_records('Error running configuration')

        encountered_error = error

    if not encountered_error:
        return

    try:
        command.execute_hooks(
            command.filter_hooks(
                config.get('commands'),
                after='error',
                action_names=arguments.keys(),
                state_names=['fail'],
            ),
            config.get('umask'),
            borgmatic.config.paths.get_working_directory(config),
            global_arguments.dry_run,
            configuration_filename=config_filename,
            log_file=config.get('log_file', ''),
            repository=error_repository.get('path', '') if error_repository else '',
            repository_label=error_repository.get('label', '') if error_repository else '',
            error=encountered_error,
            output=getattr(encountered_error, 'output', ''),
        )
    except (OSError, CalledProcessError) as error:
        if command.considered_soft_failure(error):
            return

        yield from log_error_records(f'{config_filename}: Error running after error hook', error)


def run_actions(  # noqa: PLR0912, PLR0915
    *,
    arguments,
    config_filename,
    config,
    config_paths,
    local_path,
    remote_path,
    local_borg_version,
    repository,
):
    '''
    Given parsed command-line arguments as an argparse.ArgumentParser instance, the configuration
    filename, a configuration dict, a sequence of loaded configuration paths, local and remote paths
    to Borg, a local Borg version string, and a repository dict, run all actions from the
    command-line arguments on the given repository.

    Yield JSON output strings from executing any actions that produce JSON.

    Raise OSError or subprocess.CalledProcessError if an error occurs running a command for an
    action or a hook. Raise ValueError if the arguments or configuration passed to action are
    invalid.
    '''
    add_custom_log_levels()
    repository_path = os.path.expanduser(repository['path'])
    global_arguments = arguments['global']
    dry_run_label = ' (dry run; not making any changes)' if global_arguments.dry_run else ''
    hook_context = {
        'configuration_filename': config_filename,
        'repository_label': repository.get('label', ''),
        'log_file': config.get('log_file', ''),
        # Deprecated: For backwards compatibility with borgmatic < 1.6.0.
        'repositories': ','.join([repo['path'] for repo in config['repositories']]),
        'repository': repository_path,
    }
    skip_actions = set(get_skip_actions(config, arguments))
    requested_repository = next(
        (
            repository
            for action_arguments in arguments.values()
            for repository in (getattr(action_arguments, 'repository', None),)
            if repository is not None
        ),
        None,
    )

    if requested_repository and not borgmatic.config.validate.repositories_match(
        repository,
        requested_repository,
    ):
        logger.debug('Skipping actions because the requested --repository does not match')
        return

    with borgmatic.hooks.command.Before_after_hooks(
        command_hooks=config.get('commands'),
        before_after='repository',
        umask=config.get('umask'),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        dry_run=global_arguments.dry_run,
        action_names=arguments.keys(),
        **hook_context,
    ):
        for action_name, action_arguments in arguments.items():
            if action_name == 'global' or action_name in skip_actions:
                continue

            with borgmatic.hooks.command.Before_after_hooks(
                command_hooks=config.get('commands'),
                before_after='action',
                umask=config.get('umask'),
                working_directory=borgmatic.config.paths.get_working_directory(config),
                dry_run=global_arguments.dry_run,
                action_names=(action_name,),
                **hook_context,
            ):
                if action_name == 'repo-create':
                    borgmatic.actions.repo_create.run_repo_create(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'transfer':
                    borgmatic.actions.transfer.run_transfer(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'create':
                    yield from borgmatic.actions.create.run_create(
                        config_filename,
                        repository,
                        config,
                        config_paths,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        dry_run_label,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'recreate':
                    borgmatic.actions.recreate.run_recreate(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        dry_run_label,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'prune':
                    borgmatic.actions.prune.run_prune(
                        config_filename,
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        dry_run_label,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'compact':
                    borgmatic.actions.compact.run_compact(
                        config_filename,
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        dry_run_label,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'check':
                    if checks.repository_enabled_for_checks(repository, config):
                        borgmatic.actions.check.run_check(
                            config_filename,
                            repository,
                            config,
                            local_borg_version,
                            action_arguments,
                            global_arguments,
                            local_path,
                            remote_path,
                        )
                elif action_name == 'extract':
                    borgmatic.actions.extract.run_extract(
                        config_filename,
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'export-tar':
                    borgmatic.actions.export_tar.run_export_tar(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'mount':
                    borgmatic.actions.mount.run_mount(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'restore':
                    borgmatic.actions.restore.run_restore(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'repo-list':
                    yield from borgmatic.actions.repo_list.run_repo_list(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'list':
                    yield from borgmatic.actions.list.run_list(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'repo-info':
                    yield from borgmatic.actions.repo_info.run_repo_info(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'info':
                    yield from borgmatic.actions.info.run_info(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'break-lock':
                    borgmatic.actions.break_lock.run_break_lock(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'export':
                    borgmatic.actions.export_key.run_export_key(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'import':
                    borgmatic.actions.import_key.run_import_key(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'change-passphrase':
                    borgmatic.actions.change_passphrase.run_change_passphrase(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'delete':
                    borgmatic.actions.delete.run_delete(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'repo-delete':
                    borgmatic.actions.repo_delete.run_repo_delete(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'diff':
                    borgmatic.actions.diff.run_diff(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )
                elif action_name == 'borg':
                    borgmatic.actions.borg.run_borg(
                        repository,
                        config,
                        local_borg_version,
                        action_arguments,
                        global_arguments,
                        local_path,
                        remote_path,
                    )


def load_configurations(config_filenames, arguments, overrides=None, resolve_env=True):
    '''
    Given a sequence of configuration filenames, arguments as a dict from action name to
    argparse.Namespace, a sequence of configuration file override strings in the form of
    "option.suboption=value", and whether to resolve environment variables, load and validate each
    configuration file. Return the results as a tuple of: dict of configuration filename to
    corresponding parsed configuration, a sequence of paths for all loaded configuration files
    (including includes), and a sequence of logging.LogRecord instances containing any parse errors.

    Log records are returned here instead of being logged directly because logging isn't yet
    initialized at this point! (Although with the Delayed_logging_handler now in place, maybe this
    approach could change.)
    '''
    # Dict mapping from config filename to corresponding parsed config dict.
    configs = collections.OrderedDict()
    config_paths = set()
    logs = []

    # Parse and load each configuration file.
    for config_filename in config_filenames:
        logs.extend(
            [
                logging.makeLogRecord(
                    dict(
                        levelno=logging.DEBUG,
                        levelname='DEBUG',
                        msg=f'{config_filename}: Loading configuration file',
                        name=logger.name,
                    ),
                ),
            ],
        )
        try:
            configs[config_filename], paths, parse_logs = validate.parse_configuration(
                config_filename,
                validate.schema_filename(),
                arguments,
                overrides,
                resolve_env,
            )
            config_paths.update(paths)
            logs.extend(parse_logs)
        except PermissionError:
            logs.extend(
                [
                    logging.makeLogRecord(
                        dict(
                            levelno=logging.CRITICAL,
                            levelname='CRITICAL',
                            msg=f'{config_filename}: Insufficient permissions to read configuration file',
                            name=logger.name,
                        ),
                    ),
                ],
            )
        except (ValueError, OSError, validate.Validation_error) as error:
            logs.extend(
                [
                    logging.makeLogRecord(
                        dict(
                            levelno=logging.CRITICAL,
                            levelname='CRITICAL',
                            msg=f'{config_filename}: Error parsing configuration file',
                            name=logger.name,
                        ),
                    ),
                    logging.makeLogRecord(
                        dict(
                            levelno=logging.CRITICAL,
                            levelname='CRITICAL',
                            msg=str(error),
                            name=logger.name,
                        ),
                    ),
                ],
            )

    return (configs, sorted(config_paths), logs)


def log_record(suppress_log=False, **kwargs):
    '''
    Create a log record based on the given makeLogRecord() arguments, one of which must be
    named "levelno". Log the record (unless suppress log is set) and return it.
    '''
    record = logging.makeLogRecord(dict(kwargs, name=logger.name))
    if suppress_log:
        return record

    logger.handle(record)
    return record


BORG_REPOSITORY_ACCESS_ABORTED_EXIT_CODE = 62


def log_error_records(
    message,
    error=None,
    levelno=logging.CRITICAL,
    log_command_error_output=False,
):
    '''
    Given error message text, an optional exception object, an optional log level, and whether to
    log the error output of a CalledProcessError (if any), log error summary information and also
    yield it as a series of logging.LogRecord instances.

    Note that because the logs are yielded as a generator, logs won't get logged unless you consume
    the generator output.
    '''
    level_name = logging._levelToName[levelno]

    if not error:
        yield log_record(levelno=levelno, levelname=level_name, msg=str(message))
        return

    try:
        raise error
    except CalledProcessError as called_process_error:
        yield log_record(levelno=levelno, levelname=level_name, msg=str(message))

        if called_process_error.output:
            try:
                output = called_process_error.output.decode('utf-8')
            except (UnicodeDecodeError, AttributeError):
                output = called_process_error.output

            # Suppress these logs for now and save the error output for the log summary at the end.
            # Log a separate record per line, as some errors can be really verbose and overflow the
            # per-record size limits imposed by some logging backends.
            for output_line in output.splitlines():
                yield log_record(
                    levelno=levelno,
                    levelname=level_name,
                    msg=output_line,
                    suppress_log=True,
                )

        yield log_record(levelno=levelno, levelname=level_name, msg=str(called_process_error))

        if called_process_error.returncode == BORG_REPOSITORY_ACCESS_ABORTED_EXIT_CODE:
            yield log_record(
                levelno=levelno,
                levelname=level_name,
                msg='\nTo work around this, set either the "relocated_repo_access_is_ok" or "unknown_unencrypted_repo_access_is_ok" option to "true", as appropriate.',
            )
    except (ValueError, OSError) as other_error:
        yield log_record(levelno=levelno, levelname=level_name, msg=str(message))
        yield log_record(levelno=levelno, levelname=level_name, msg=str(other_error))
    except:  # noqa: E722, S110
        # Raising above only as a means of determining the error type. Swallow the exception here
        # because we don't want the exception to propagate out of this function.
        pass


def get_local_path(configs):
    '''
    Arbitrarily return the local path from the first configuration dict. Default to "borg" if not
    set.
    '''
    return next(iter(configs.values())).get('local_path', 'borg')


def collect_highlander_action_summary_logs(configs, arguments, configuration_parse_errors):
    '''
    Given a dict of configuration filename to corresponding parsed configuration, parsed
    command-line arguments as a dict from subparser name to a parsed namespace of arguments, and
    whether any configuration files encountered errors during parsing, run a highlander action
    specified in the arguments, if any, and yield a series of logging.LogRecord instances containing
    summary information.

    A highlander action is an action that cannot coexist with other actions on the borgmatic
    command-line, and borgmatic exits after processing such an action.
    '''
    add_custom_log_levels()

    try:
        if 'bootstrap' in arguments:
            try:
                # No configuration file is needed for bootstrap.
                local_borg_version = borg_version.local_borg_version(
                    {},
                    arguments['bootstrap'].local_path,
                )
            except (OSError, CalledProcessError, ValueError) as error:
                yield from log_error_records('Error getting local Borg version', error)
                return

            borgmatic.actions.config.bootstrap.run_bootstrap(
                arguments['bootstrap'],
                arguments['global'],
                local_borg_version,
            )
            yield logging.makeLogRecord(
                dict(
                    levelno=logging.ANSWER,
                    levelname='ANSWER',
                    msg='Bootstrap successful',
                    name=logger.name,
                ),
            )

            return

        if 'generate' in arguments:
            borgmatic.actions.config.generate.run_generate(
                arguments['generate'],
                arguments['global'],
            )
            yield logging.makeLogRecord(
                dict(
                    levelno=logging.ANSWER,
                    levelname='ANSWER',
                    msg='Generate successful',
                    name=logger.name,
                ),
            )

            return

        if 'validate' in arguments:
            if configuration_parse_errors:
                yield logging.makeLogRecord(
                    dict(
                        levelno=logging.CRITICAL,
                        levelname='CRITICAL',
                        msg='Configuration validation failed',
                        name=logger.name,
                    ),
                )

                return

            borgmatic.actions.config.validate.run_validate(arguments['validate'], configs)

            yield logging.makeLogRecord(
                dict(
                    levelno=logging.ANSWER,
                    levelname='ANSWER',
                    msg='All configuration files are valid',
                    name=logger.name,
                ),
            )

            return

        if 'show' in arguments:
            borgmatic.actions.config.show.run_show(arguments['show'], configs)

            return

    except (
        CalledProcessError,
        ValueError,
        OSError,
    ) as error:
        yield from log_error_records(error)


def collect_configuration_run_summary_logs(configs, config_paths, arguments, log_file_path):  # noqa: PLR0912
    '''
    Given a dict of configuration filename to corresponding parsed configuration, a sequence of
    loaded configuration paths, parsed command-line arguments as a dict from subparser name to a
    parsed namespace of arguments, and the path of a log file (if any), run each configuration file
    and yield a series of logging.LogRecord instances containing summary information about each run.

    As a side effect of running through these configuration files, output their JSON results, if
    any, to stdout.
    '''
    # Run cross-file validation checks.
    repository = None

    for action_arguments in arguments.values():
        if hasattr(action_arguments, 'repository'):
            repository = action_arguments.repository
            break

    try:
        validate.guard_configuration_contains_repository(repository, configs)
    except ValueError as error:
        yield from log_error_records(str(error))
        return

    if not configs:
        yield from log_error_records(
            f"{' '.join(arguments['global'].config_paths)}: No valid configuration files found",
        )
        return

    try:
        seen_command_hooks = []

        for config_filename, config in configs.items():
            command_hooks = command.filter_hooks(
                tuple(
                    command_hook
                    for command_hook in config.get('commands', ())
                    if command_hook not in seen_command_hooks
                ),
                before='everything',
                action_names=arguments.keys(),
            )

            if command_hooks:
                command.execute_hooks(
                    command_hooks,
                    config.get('umask'),
                    borgmatic.config.paths.get_working_directory(config),
                    arguments['global'].dry_run,
                    configuration_filename=config_filename,
                    log_file=log_file_path or '',
                )
                seen_command_hooks += list(command_hooks)
    except (CalledProcessError, ValueError, OSError) as error:
        yield from log_error_records('Error running before everything hook', error)
        return

    # Execute the actions corresponding to each configuration file.
    json_results = []
    encountered_error = False

    for config_filename, config in configs.items():
        with Log_prefix(config_filename):
            results = list(run_configuration(config_filename, config, config_paths, arguments))

            error_logs = tuple(
                result for result in results if isinstance(result, logging.LogRecord)
            )

            if error_logs:
                encountered_error = True
                yield from log_error_records('An error occurred')
                yield from error_logs
            else:
                yield logging.makeLogRecord(
                    dict(
                        levelno=logging.INFO,
                        levelname='INFO',
                        msg=f'{config_filename}: Successfully ran configuration file',
                        name=logger.name,
                    ),
                )
                if results:
                    json_results.extend(results)

    if 'umount' in arguments:
        logger.info(f"Unmounting mount point {arguments['umount'].mount_point}")
        try:
            borg_umount.unmount_archive(
                config,
                mount_point=arguments['umount'].mount_point,
                local_path=get_local_path(configs),
            )
        except (CalledProcessError, OSError) as error:
            encountered_error = True
            yield from log_error_records('Error unmounting mount point', error)

    if json_results:
        sys.stdout.write(json.dumps(json_results))

    try:
        seen_command_hooks = []

        for config_filename, config in configs.items():
            command_hooks = command.filter_hooks(
                tuple(
                    command_hook
                    for command_hook in config.get('commands', ())
                    if command_hook not in seen_command_hooks
                ),
                after='everything',
                action_names=arguments.keys(),
                state_names=['fail' if encountered_error else 'finish'],
            )

            if command_hooks:
                command.execute_hooks(
                    command_hooks,
                    config.get('umask'),
                    borgmatic.config.paths.get_working_directory(config),
                    arguments['global'].dry_run,
                    configuration_filename=config_filename,
                    log_file=log_file_path or '',
                )
                seen_command_hooks += list(command_hooks)
    except (CalledProcessError, ValueError, OSError) as error:
        yield from log_error_records('Error running after everything hook', error)


def exit_with_help_link():  # pragma: no cover
    '''
    Display a link to get help and exit with an error code.
    '''
    logger.critical('')
    logger.critical('Need some help? https://torsion.org/borgmatic/#issues')
    sys.exit(1)


def check_and_show_help_on_no_args(configs, schema):
    '''
    Given a dict of configuration filename to corresponding parsed configuration and the
    configuration schema as a dict, check if the borgmatic command was run without any arguments. If
    the configuration option "default_actions" is set to False, then show the help message an exit.
    '''
    if len(sys.argv) == 1:  # No arguments provided.
        default_actions = any(config.get('default_actions', True) for config in configs.values())

        if configs and not default_actions:
            parse_arguments(schema, '--help')
            sys.exit(0)


def get_singular_option_value(configs, option_name):
    '''
    Given a dict of configuration filename to corresponding parsed configuration, return the value
    of the given option from the configuration files or None if it's not set.

    Log and exit if there are conflicting values for the option across the configuration files.
    '''
    distinct_values = {
        value
        for config in configs.values()
        for value in (config.get(option_name),)
        if value is not None
    }

    if len(distinct_values) > 1:
        configure_logging(logging.CRITICAL)
        joined_values = ', '.join(str(value) for value in distinct_values)
        logger.critical(
            f'The {option_name} option has conflicting values across configuration files: {joined_values}',
        )
        exit_with_help_link()

    try:
        return next(iter(distinct_values))
    except StopIteration:
        return None


def display_summary(summary_logs, log_json):  # pragma: no cover
    summary_logs_max_level = max(log.levelno for log in summary_logs)

    for message in ('summary:',) if log_json else ('', 'summary:'):
        log_record(
            levelno=summary_logs_max_level,
            levelname=logging.getLevelName(summary_logs_max_level),
            msg=message,
        )

    for log in summary_logs:
        logger.handle(log)

    if summary_logs_max_level >= logging.CRITICAL:
        exit_with_help_link()


def main(extra_summary_logs=()):  # pragma: no cover
    configure_signals()
    configure_delayed_logging()
    schema_filename = validate.schema_filename()

    try:
        schema = borgmatic.config.load.load_configuration(schema_filename)
    except (ruamel.yaml.error.YAMLError, RecursionError) as error:
        configure_logging(logging.CRITICAL)
        logger.critical(error)
        exit_with_help_link()

    try:
        arguments = parse_arguments(schema, *sys.argv[1:])
    except ValueError as error:
        configure_logging(logging.CRITICAL)
        logger.critical(error)
        exit_with_help_link()
    except SystemExit as error:
        if error.code == 0:
            raise

        configure_logging(logging.CRITICAL)
        logger.critical(f"Error parsing arguments: {' '.join(sys.argv)}")
        exit_with_help_link()

    global_arguments = arguments['global']

    if global_arguments.version:
        print(importlib.metadata.version('borgmatic'))  # noqa: T201
        sys.exit(0)

    if global_arguments.bash_completion:
        print(borgmatic.commands.completion.bash.bash_completion())  # noqa: T201
        sys.exit(0)

    if global_arguments.fish_completion:
        print(borgmatic.commands.completion.fish.fish_completion())  # noqa: T201
        sys.exit(0)

    config_filenames = tuple(collect.collect_config_filenames(global_arguments.config_paths))
    configs, config_paths, parse_logs = load_configurations(
        config_filenames,
        arguments,
        global_arguments.overrides,
        resolve_env=global_arguments.resolve_env and not arguments.get('validate'),
    )

    check_and_show_help_on_no_args(configs, schema)

    configuration_parse_errors = (
        (max(log.levelno for log in parse_logs) >= logging.CRITICAL) if parse_logs else False
    )

    any_json_flags = any(
        getattr(sub_arguments, 'json', False) for sub_arguments in arguments.values()
    )
    log_file_path = get_singular_option_value(configs, 'log_file')
    log_json = get_singular_option_value(configs, 'log_json')

    try:
        configure_logging(
            verbosity_to_log_level(get_verbosity(configs, 'verbosity')),
            verbosity_to_log_level(get_verbosity(configs, 'syslog_verbosity')),
            verbosity_to_log_level(get_verbosity(configs, 'log_file_verbosity')),
            verbosity_to_log_level(get_verbosity(configs, 'monitoring_verbosity')),
            log_file_path,
            get_singular_option_value(configs, 'log_file_format'),
            log_json,
            color_enabled=should_do_markup(configs, any_json_flags or log_json),
        )
    except (FileNotFoundError, PermissionError) as error:
        configure_logging(logging.CRITICAL)
        logger.critical(f'Error configuring logging: {error}')
        exit_with_help_link()

    summary_logs = (
        list(extra_summary_logs)
        + parse_logs
        + (
            list(
                collect_highlander_action_summary_logs(
                    configs,
                    arguments,
                    configuration_parse_errors,
                ),
            )
            or list(
                collect_configuration_run_summary_logs(
                    configs,
                    config_paths,
                    arguments,
                    log_file_path,
                ),
            )
        )
    )

    display_summary(summary_logs, log_json)


if __name__ == '__main__':  # pragma: no cover
    main()



import io
import re

import ruamel.yaml

import borgmatic.config.schema

LIST_INDEX_KEY_PATTERN = re.compile(r'^(?P<list_name>[a-zA-z-]+)\[(?P<index>\d+)\]$')


def set_values(config, keys, value):
    '''
    Given a configuration dict, a sequence of parsed key strings, and a string value, descend into
    the configuration hierarchy based on the given keys and set the value into the right place.
    For example, consider these keys:

        ('foo', 'bar', 'baz')

    This looks up "foo" in the given configuration dict. And within that, it looks up "bar". And
    then within that, it looks up "baz" and sets it to the given value. Another example:

        ('mylist[0]', 'foo')

    This looks for the zeroth element of "mylist" in the given configuration. And within that, it
    looks up "foo" and sets it to the given value.
    '''
    if not keys:
        return

    first_key = keys[0]

    # Support "mylist[0]" list index syntax.
    match = LIST_INDEX_KEY_PATTERN.match(first_key)

    if match:
        list_key = match.group('list_name')
        list_index = int(match.group('index'))

        try:
            if len(keys) == 1:
                config[list_key][list_index] = value

                return

            if list_key not in config:
                config[list_key] = []

            set_values(config[list_key][list_index], keys[1:], value)
        except (IndexError, KeyError):
            raise ValueError(f'Argument list index {first_key} is out of range')

        return

    if len(keys) == 1:
        config[first_key] = value

        return

    if first_key not in config:
        config[first_key] = {}

    set_values(config[first_key], keys[1:], value)


def type_for_option(schema, option_keys):
    '''
    Given a configuration schema dict and a sequence of keys identifying a potentially nested
    option, e.g. ('extra_borg_options', 'create'), return the schema type of that option as a
    string.

    Return None if the option or its type cannot be found in the schema.
    '''
    option_schema = schema

    for key in option_keys:
        # Support "name[0]"-style list index syntax.
        match = LIST_INDEX_KEY_PATTERN.match(key)
        properties = borgmatic.config.schema.get_properties(option_schema)

        try:
            if match:
                option_schema = properties[match.group('list_name')]['items']
            else:
                option_schema = properties[key]
        except KeyError:
            return None

    try:
        return option_schema['type']
    except KeyError:
        return None


def convert_value_type(value, option_type):
    '''
    Given a string value and its schema type as a string, determine its logical type (string,
    boolean, integer, etc.), and return it converted to that type.

    If the destination option type is a string, then leave the value as-is so that special
    characters in it don't get interpreted as YAML during conversion.

    And if the source value isn't a string, return it as-is.

    Raise ruamel.yaml.error.YAMLError if there's a parse issue with the YAML.
    Raise ValueError if the parsed value doesn't match the option type.
    '''
    if not isinstance(value, str):
        return value

    if option_type == 'string':
        return value

    try:
        parsed_value = ruamel.yaml.YAML(typ='safe').load(io.StringIO(value))
    except ruamel.yaml.error.YAMLError as error:
        raise ValueError(f'Argument value "{value}" is invalid: {error.problem}')

    if not isinstance(parsed_value, borgmatic.config.schema.parse_type(option_type)):
        raise ValueError(f'Argument value "{value}" is not of the expected type: {option_type}')

    return parsed_value


def prepare_arguments_for_config(global_arguments, schema):
    '''
    Given global arguments as an argparse.Namespace and a configuration schema dict, parse each
    argument that corresponds to an option in the schema and return a sequence of tuples (keys,
    values) for that option, where keys is a sequence of strings. For instance, given the following
    arguments:

        argparse.Namespace(**{'my_option.sub_option': 'value1', 'other_option': 'value2'})

    ... return this:

        (
            (('my_option', 'sub_option'), 'value1'),
            (('other_option',), 'value2'),
        )
    '''
    prepared_values = []

    for argument_name, value in global_arguments.__dict__.items():
        if value is None:
            continue

        keys = tuple(argument_name.split('.'))
        option_type = type_for_option(schema, keys)

        # The argument doesn't correspond to any option in the schema, or it is a complex argument, so ignore it.
        # It's probably a flag that borgmatic has on the command-line but not in configuration.
        if option_type in {'object', None}:
            continue

        prepared_values.append(
            (
                keys,
                convert_value_type(value, option_type),
            ),
        )

    return tuple(prepared_values)


def apply_arguments_to_config(config, schema, arguments):
    '''
    Given a configuration dict, a corresponding configuration schema dict, and arguments as a dict
    from action name to argparse.Namespace, set those given argument values into their corresponding
    configuration options in the configuration dict.

    This supports argument flags of the from "--foo.bar.baz" where each dotted component is a nested
    configuration object. Additionally, flags like "--foo.bar[0].baz" are supported to update a list
    element in the configuration.
    '''
    for action_arguments in arguments.values():
        for keys, value in prepare_arguments_for_config(action_arguments, schema):
            set_values(config, keys, value)



import functools
import logging
import shlex

import borgmatic.config.paths
import borgmatic.execute

logger = logging.getLogger(__name__)


@functools.cache
def run_passcommand(passcommand, working_directory):
    '''
    Run the given passcommand using the given working directory and return the passphrase produced
    by the command.

    Cache the results so that the passcommand only needs to run—and potentially prompt the user—once
    per borgmatic invocation.
    '''
    return '\n'.join(
        borgmatic.execute.execute_command_and_capture_output(
            shlex.split(passcommand),
            working_directory=working_directory,
        )
    )


def get_passphrase_from_passcommand(config):
    '''
    Given the configuration dict, call the configured passcommand to produce and return an
    encryption passphrase. In effect, we're doing an end-run around Borg by invoking its passcommand
    ourselves. This allows us to pass the resulting passphrase to multiple different Borg
    invocations without the user having to be prompted multiple times.

    If no passcommand is configured, then return None.
    '''
    passcommand = config.get('encryption_passcommand')

    if not passcommand:
        return None

    return run_passcommand(passcommand, borgmatic.config.paths.get_working_directory(config))



import logging
import shlex

import borgmatic.config.paths
from borgmatic.borg import environment, feature, flags
from borgmatic.execute import execute_command

logger = logging.getLogger(__name__)


def compact_segments(
    dry_run,
    repository_path,
    config,
    local_borg_version,
    global_arguments,
    local_path='borg',
    remote_path=None,
    cleanup_commits=False,
):
    '''
    Given dry-run flag, a local or remote repository path, a configuration dict, and the local Borg
    version, compact the segments in a repository.
    '''
    umask = config.get('umask', None)
    lock_wait = config.get('lock_wait', None)
    extra_borg_options = config.get('extra_borg_options', {}).get('compact', '')
    threshold = config.get('compact_threshold')

    full_command = (
        (local_path, 'compact')
        + (('--remote-path', remote_path) if remote_path else ())
        + (('--umask', str(umask)) if umask else ())
        + (('--log-json',) if (config.get('log_json') or not config.get('progress')) else ())
        + (('--lock-wait', str(lock_wait)) if lock_wait else ())
        + (('--progress',) if config.get('progress') else ())
        + (('--cleanup-commits',) if cleanup_commits else ())
        + (('--threshold', str(threshold)) if threshold else ())
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + (
            ('--dry-run',)
            if dry_run and feature.available(feature.Feature.DRY_RUN_COMPACT, local_borg_version)
            else ()
        )
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_flags(repository_path, local_borg_version)
    )

    if dry_run and not feature.available(feature.Feature.DRY_RUN_COMPACT, local_borg_version):
        logger.info('Skipping compact (dry run)')
        return

    execute_command(
        full_command,
        output_log_level=logging.INFO,
        environment=environment.make_environment(config),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import logging
import os
import re
import shlex

import borgmatic.execute

logger = logging.getLogger(__name__)


CREDENTIAL_NAME_PATTERN = re.compile(r'^[\w.-]+$')


def load_credential(hook_config, config, credential_parameters):
    '''
    Given the hook configuration dict, the configuration dict, and a credential parameters tuple
    containing a credential name to load, read the credential from the corresponding systemd
    credential file and return it.

    Raise ValueError if the systemd CREDENTIALS_DIRECTORY environment variable is not set, the
    credential name is invalid, or the credential file cannot be read.
    '''
    try:
        (credential_name,) = credential_parameters
    except ValueError:
        name = ' '.join(credential_parameters)

        raise ValueError(f'Cannot load invalid credential name: "{name}"')

    if not CREDENTIAL_NAME_PATTERN.match(credential_name):
        raise ValueError(f'Cannot load invalid credential name "{credential_name}"')

    credentials_directory = os.environ.get('CREDENTIALS_DIRECTORY')

    if not credentials_directory:
        logger.debug(
            f'Falling back to loading credential "{credential_name}" via systemd-creds because the systemd CREDENTIALS_DIRECTORY environment variable is not set'
        )

        command = (
            *shlex.split((hook_config or {}).get('systemd_creds_command', 'systemd-creds')),
            'decrypt',
            os.path.join(
                (hook_config or {}).get(
                    'encrypted_credentials_directory', '/etc/credstore.encrypted'
                ),
                credential_name,
            ),
        )

        return '\n'.join(borgmatic.execute.execute_command_and_capture_output(command)).rstrip(
            os.linesep
        )

    try:
        with open(
            os.path.join(credentials_directory, credential_name), encoding='utf-8'
        ) as credential_file:
            return credential_file.read().rstrip(os.linesep)
    except (FileNotFoundError, OSError) as error:
        logger.warning(error)

        raise ValueError(f'Cannot load credential "{credential_name}" from file: {error.filename}')
